package org.openl.itest;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.concurrent.TimeUnit;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.openl.itest.core.HttpClient;
import org.openl.itest.core.JettyServer;
import org.openl.itest.core.worker.AsyncExecutor;
import org.openl.itest.core.worker.TaskScheduler;
import org.openl.itest.response.ServiceInfoResponse;
import org.openl.itest.response.UiInfoResponse;

public class RunRestRulesDeploymentTest {

    private static JettyServer server;
    private static HttpClient client;

    @BeforeClass
    public static void setUp() throws Exception {
        server = JettyServer.start();
        client = server.client();
    }

    @AfterClass
    public static void tearDown() throws Exception {
        server.stop();
    }

    @Test
    public void testDeployRules() {
        client.get("/admin/services", "/no_services.resp.txt");
        client.post("/REST/deployed-rules/hello", "/deployed-rules_hello.req.json", 404, "/404.html");
        long createServiceTime = System.currentTimeMillis();
        client.post("/admin/deploy", "/rules-to-deploy.zip", 201);

        ServiceInfoResponse[] servicesInfo2 = client.get("/admin/services", ServiceInfoResponse[].class);
        assertEquals(1, servicesInfo2.length);
        checkServiceInfo(servicesInfo2[0], "deployed-rules", "deployed-rules", "REST/deployed-rules");
        long service2Time = servicesInfo2[0].getStartedTime().getTime();
        checkServiceTime(service2Time, createServiceTime);

        client.get("/admin/read/deployed-rules", "/rules-to-deploy.zip");
        client.get("/admin/services/deployed-rules/methods", "/deployed-rules_methods.resp.txt");

        client.post("/REST/deployed-rules/hello", "/deployed-rules_hello.req.json", "/deployed-rules_hello.resp.txt");
        // should not be updated
        client.post("/admin/deploy", "/rules-to-deploy_v2.zip", 201);

        ServiceInfoResponse[] servicesInfo3 = client.get("/admin/services", ServiceInfoResponse[].class);
        assertEquals(1, servicesInfo3.length);
        checkServiceInfo(servicesInfo3[0], "deployed-rules", "deployed-rules", "REST/deployed-rules");
        long service3Time = servicesInfo3[0].getStartedTime().getTime();
        assertEquals(service2Time, service3Time);

        client.post("/REST/deployed-rules/hello", "/deployed-rules_hello.req.json", "/deployed-rules_hello.resp.txt");

        // should be updated
        createServiceTime = System.currentTimeMillis();
        client.put("/admin/deploy", "/rules-to-deploy_v2.zip", 201);

        ServiceInfoResponse[] servicesInfo4 = client.get("/admin/services", ServiceInfoResponse[].class);
        assertEquals(1, servicesInfo4.length);
        checkServiceInfo(servicesInfo4[0], "deployed-rules", "deployed-rules", "REST/deployed-rules");
        checkServiceTime(servicesInfo4[0].getStartedTime().getTime(), createServiceTime);

        UiInfoResponse uiInfoResponseResponseEntity = client.get("/admin/ui/info", UiInfoResponse.class);
        ServiceInfoResponse[] serviceInfo = uiInfoResponseResponseEntity.getServices();
        assertTrue(uiInfoResponseResponseEntity.getDeployerEnabled());
        assertEquals(1, serviceInfo.length);
        assertEquals("deployed-rules", serviceInfo[0].getName());
        assertEquals(ServiceInfoResponse.ServiceStatus.DEPLOYED, serviceInfo[0].getStatus());

        client.post("/REST/deployed-rules/hello", "/deployed-rules_hello.req.json", "/deployed-rules_hello_2.resp.txt");

        client.put("/admin/deploy", "/rules-to-deploy-failed.zip", 201);
        ServiceInfoResponse[] servicesInfo5 = client.get("/admin/services", ServiceInfoResponse[].class);
        assertEquals(1, servicesInfo5.length);
        assertEquals(ServiceInfoResponse.ServiceStatus.FAILED, servicesInfo5[0].getStatus());
        client.get("/admin/services/deployed-rules/errors", "/deployed-rules_errors.resp.json");

        client.get("/admin/services/deployed-rules/MANIFEST.MF", 404);
        client.delete("/admin/delete/deployed-rules");
        UiInfoResponse uiInfoResponseResponseEntity2 = client.get("/admin/ui/info", UiInfoResponse.class);
        assertEquals(0, uiInfoResponseResponseEntity2.getServices().length);
        assertTrue(uiInfoResponseResponseEntity.getDeployerEnabled());

        client.post("/admin/deploy", "/empty_project.zip", 400);
    }

    @Test
    public void testDeployRules_multipleDeployment() {
        client.get("/admin/services", "/no_services.resp.txt");
        client.post("/REST/project1/sayHello", "/project1_sayHello.req.txt", 404, "/404.html");

        long createServiceTime = System.currentTimeMillis();
        client.post("/admin/deploy", "/multiple-deployment_v1.zip", 201);

        ServiceInfoResponse[] servicesInfo = client.get("/admin/services", ServiceInfoResponse[].class);
        assertEquals(2, servicesInfo.length);
        checkServiceInfo(servicesInfo[0], "project1", "project1", "REST/project1");
        checkServiceTime(servicesInfo[0].getStartedTime().getTime(), createServiceTime);
        checkServiceInfo(servicesInfo[1],
            "yaml_project_project2",
            "yaml_project/project2",
            "REST/yaml_project/project2");
        checkServiceTime(servicesInfo[1].getStartedTime().getTime(), createServiceTime);

        client.get("/admin/services/project1/methods", "/project1_methods.resp.txt");
        client.get("/admin/services/yaml_project_project2/methods", "/yaml_project_project2_methods.resp.txt");

        client.post("/REST/project1/sayHello", "/project1_sayHello.req.txt", "/project1_sayHello.resp.txt");

        // should not be updated
        client.post("/admin/deploy", "/multiple-deployment_v2.zip", 201);
        client.post("/REST/project1/sayHello", "/project1_sayHello.req.txt", "/project1_sayHello.resp.txt");

        // should be updated
        client.put("/admin/deploy", "/multiple-deployment_v2.zip", 201);
        client.post("/REST/project1/sayHello", "/project1_sayHello.req.txt", "/project1_sayHello_2.resp.txt");

        client.delete("/admin/delete/project1");
        client.delete("/admin/delete/yaml_project_project2");
    }

    @Test
    public void testMissingServiceMethods() {
        client.get("/admin/services/missing-name/methods", 404);
        client.delete("/admin/delete/missing-name", 404);
        client.get("/admin/read/missing-name", 404);
        client.get("/admin/services/missing-name/errors", 404);
        client.get("/admin/services/missing-name/MANIFEST.MF", 404);
    }

    @Test
    public void test_EPBDS_8758_multithread() {
        client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v1.zip", 201);
        client.get("/REST/EPBDS-8758/doSomething", "/EPBDS-8758/doSomething_v1.resp.txt");
        AsyncExecutor executor = new AsyncExecutor(() -> client.get("/REST/EPBDS-8758/doSomething", String.class));
        TaskScheduler taskScheduler = new TaskScheduler();

        executor.start();

        taskScheduler.schedule(() -> {
            client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v2.zip", 201);
            client.get("/REST/EPBDS-8758/doSomething", "/EPBDS-8758/doSomething_v2.resp.txt");
        }, 1, TimeUnit.SECONDS);
        taskScheduler.schedule(() -> {
            client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v3.zip", 201);
            client.get("/REST/EPBDS-8758/doSomething", "/EPBDS-8758/doSomething_v3.resp.txt");
        }, 2, TimeUnit.SECONDS);

        boolean deployErrors = taskScheduler.await();
        boolean invocationErrors = executor.stop();

        client.delete("/admin/delete/EPBDS-8758");

        assertFalse(invocationErrors);
        assertFalse(deployErrors);
    }

    @Test
    public void test_EPBDS_8758_multithread2() throws Exception {
        client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v1.zip", 201);
        client.get("/REST/EPBDS-8758/doSomething", "/EPBDS-8758/doSomething_v1.resp.txt");
        AsyncExecutor executor = new AsyncExecutor(() -> client.send("EPBDS-8758/doSomething.get"));
        executor.start();

        AsyncExecutor deployers = AsyncExecutor.start(() -> {
            client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v2.zip", 201);
        }, () -> {
            client.put("/admin/deploy", "/EPBDS-8758/EPBDS-8758-v3.zip", 201);
        });

        TimeUnit.SECONDS.sleep(1);
        boolean deployErrors = deployers.stop();
        boolean invocationErrors = executor.stop();

        client.delete("/admin/delete/EPBDS-8758");

        assertFalse(deployErrors);
        assertFalse(invocationErrors);
    }

    @Test
    public void test_EPBDS_10068_MANIFEST() {
        client.put("/admin/deploy", "/EPBDS-10068/EPBDS-10068.zip", 201);
        client.get("/admin/services/EPBDS-10068/MANIFEST.MF", "/EPBDS-10068/MANIFEST.MF.resp.json");
        client.delete("/admin/delete/EPBDS-10068");
    }

    @Test
    public void test_EPBDS_10157() {
        client.put("/admin/deploy/EPBDS-10157", "/EPBDS-10157/EPBDS-10157.zip", 201);
        client.get("/EPBDS-10157/EPBDS-10157/doSomething", "/EPBDS-10157/EPBDS-10157.resp.txt");
        client.put("/admin/deploy/EPBDS-10157_2", "/EPBDS-10157/EPBDS-10157_2.zip", 201);
        client.get("/ /doSomething", "/EPBDS-10157/EPBDS-10157.resp.txt");
        client.put("/admin/deploy/EPBDS-9902", "/EPBDS-10157/EPBDS-9902.zip", 201);
        client.get("/EPBDS-9902/EPBDS-9902/doSomething", "/EPBDS-10157/EPBDS-10157.resp.txt");
        client.delete("/admin/delete/EPBDS-10157_EPBDS-10157");
        client.delete("/admin/delete/ ");
        client.delete("/admin/delete/EPBDS-9902_EPBDS-9902");
    }

    private void checkServiceInfo(ServiceInfoResponse service,
            String expectedName,
            String expectedSoap,
            String expectedRest) {
        assertEquals(expectedName, service.getName());
        assertEquals(expectedSoap, service.getUrls().get("WEBSERVICE"));
        assertEquals(expectedRest, service.getUrls().get("RESTFUL"));
    }

    private void checkServiceTime(long timeToCheckMs, long createServiceTime) {
        assertTrue((timeToCheckMs >= createServiceTime && timeToCheckMs <= System.currentTimeMillis()));
    }

}
